using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Xml.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;

// TODO: replace this with the type you want to read.
using Ohsai;
using TRead = OhsaiContentReaders.ReaderResult;

namespace OhsaiContentReaders
{
    /// <summary>
    /// This class will be instantiated by the XNA Framework Content
    /// Pipeline to read the specified data type from binary .xnb format.
    /// 
    /// Unlike the other Content Pipeline support classes, this should
    /// be a part of your main game project, and not the Content Pipeline
    /// Extension Library project.
    /// </summary>
    public class OhsaiTypeReader : ContentTypeReader<TRead>
    {
        protected override TRead Read(ContentReader input, TRead existingInstance)
        {
            var returnValue = new ReaderResult();
            returnValue.Success = true;

            var doc = XDocument.Load(new StringReader(input.ReadString()));

            IAgent agent = this.CreateAgent(doc.Root);

            this.TryParseResponders(doc, agent);

            returnValue.Agent = agent;
            return returnValue;
        }

        private void TryParseResponders(XDocument doc, IAgent agent)
        {
            var responderNodes = from node in doc.Root.Descendants()
                                 where node.Name == "responder"
                                 select node;

            foreach (var node in responderNodes)
            {
                // get name
                var name = node.Attribute("type");

                // try to find and create type
                var responderObj = name != null ? this.FindAndCreateType<Ohsai.IResponder>(name.Value) : null;

                if (responderObj != null)
                {
                    // try to parse the conditional
                    this.TryParseConditional(node, responderObj, agent);

                    // try to parse the conditional
                    this.TryParseAction(node, responderObj, agent);

                    // try to read in responder properties
                    this.TryParseProperties(node, responderObj, agent);
                }


                agent.Responders.Add(responderObj);
            }
        }

        private IAgent CreateAgent(XElement root)
        {
            var agentNode = (from node in root.Descendants()
                             where node.Name == "agent"
                             select node).FirstOrDefault();
            if (agentNode != null)
            {
                var name = agentNode.Attribute("type");
                if (name != null)
                {
                    var agent = this.FindAndCreateType<Ohsai.IAgent>(name.Value);
                    this.TryParseProperties(agentNode, agent, agent);
                    return agent;
                }
            }

            return new BasicAgent();
        }

        private void TryParseConditional(XElement node, IResponder responderObj, IAgent agent)
        {
            // try to parse condition
            var conNode = (node.Descendants().Where(c => c.Name == "condition")).FirstOrDefault();
            if (conNode != null)
            {
                var name = conNode.Attribute("type");
                var conditionObj = name != null ? this.FindAndCreateType<ICondition>(name.Value) : null;

                if (conditionObj != null)
                {
                    responderObj.Condition = conditionObj;

                    // try to read in condition properties
                    this.TryParseProperties(conNode, conditionObj, agent);

                    // try to parse the sensor object
                    this.TryParseSensor(node, conditionObj, agent);

                    // try to parse any filters
                    this.TryParseFilter(conNode, conditionObj, agent);
                }
            }
        }

        private void TryParseAction(XElement node, IResponder responderObj, IAgent agent)
        {
            // try to parse condition
            var actionNode = (node.Descendants().Where(c => c.Name == "action")).FirstOrDefault();
            if (actionNode != null)
            {
                var name = actionNode.Attribute("type");
                var actionObj = name != null ? this.FindAndCreateType<IAction>(name.Value) : null;

                if (actionObj != null)
                {
                    responderObj.Action = actionObj;

                    // try to read in action properties
                    this.TryParseProperties(actionNode, actionObj, agent);
                    // try to parse the verb
                    this.TryParseVerb(actionNode, actionObj, agent);
                    // try to parse for any modifiers
                    this.TryParseModifiers(actionNode, actionObj, agent);
                }
            }
        }

        private void TryParseSensor(XElement node, ICondition conditionObj, IAgent agent)
        {
            // try to parse sensor
            var sensorNode = (from c in node.Descendants()
                              where c.Name == "sensor"
                              select c).FirstOrDefault();
            if (sensorNode != null)
            {
                var name = sensorNode.Attribute("type");
                var sensorObj = name != null ? this.FindAndCreateType<ISensor>(name.Value) : null;

                if (sensorObj != null)
                {
                    conditionObj.Sensor = sensorObj;

                    // try to read in properties
                    this.TryParseProperties(sensorNode, sensorObj, agent);
                }
            }
        }

        private void TryParseFilter(XElement node, ICondition conditionObj, IAgent agent)
        {
            var filterNodes = node.Descendants().Where(fn => fn.Name == "filter").OrderBy(fn => fn.Attribute("index"));
            foreach (var filterNode in filterNodes)
            {
                // try to parse filter
                var name = filterNode.Attribute("type");
                var filterObj = name != null ? this.FindAndCreateType<Ohsai.ISensorFilter>(name.Value) : null;

                if (filterObj != null)
                {
                    conditionObj.Filters.Add(filterObj);

                    // try to read in properties
                    this.TryParseProperties(filterNode, filterObj, agent);
                }
            }
        }

        private void TryParseModifiers(XElement node, IAction actionObj, IAgent agent)
        {
            var filterNodes = node.Descendants().Where(fn => fn.Name == "modifier").OrderBy(fn => fn.Attribute("index"));
            foreach (var modifierNode in filterNodes)
            {
                // try to parse modifier
                var name = modifierNode.Attribute("type");
                if (this.ModifierCanBeApplied(actionObj, name != null ? name.Value : null))
                {
                    var modfierObj = name != null ? this.FindAndCreateType<Ohsai.IModifier>(name.Value) : null;

                    if (modfierObj != null)
                    {
                        actionObj.Modifiers.Add(modfierObj);

                        // try to read in properties
                        this.TryParseProperties(modifierNode, modfierObj, agent);
                    }
                }
            }
        }

        private bool ModifierCanBeApplied(IAction actionObj, string name)
        {
            if (name == null) return false;

            var results = from asm in AppDomain.CurrentDomain.GetAssemblies()
                          where asm.GlobalAssemblyCache == false
                          from ty in asm.GetTypes()
                          let att = ty.GetCustomAttributes(true).Where(s => s is OhsaiModifier).FirstOrDefault()
                          where ty.FullName == name && ty.GetInterface(typeof(IModifier).FullName) != null &&
                                att != null && ((OhsaiModifier)att).Types.Contains(actionObj.Verb.GetType().FullName)
                          select ty;

            return results.Count() > 0;
        }

        private void TryParseVerb(XElement node, IAction actionObj, IAgent agent)
        {
            // try to parse sensor
            var verbNode = (from c in node.Descendants()
                            where c.Name == "verb"
                            select c).FirstOrDefault();
            if (verbNode != null)
            {
                var name = verbNode.Attribute("type");
                var verbObj = name != null ? this.FindAndCreateType<Ohsai.IVerb>(name.Value) : null;

                if (verbObj != null)
                {
                    actionObj.Verb = verbObj;

                    // try to read in properties
                    this.TryParseProperties(verbNode, verbObj, agent);
                }
            }
        }

        private void TryParseProperties(XElement node, object obj, IAgent agent)
        {
            // get props
            var props = from p in obj.GetType().GetProperties()
                        from att in p.GetCustomAttributes(false)
                        where att is OhsaiProperty && node.Descendants().Where(s => p.Name == s.Name).Count() > 0
                        select p;
            var fields = from p in obj.GetType().GetFields()
                         from att in p.GetCustomAttributes(false)
                         where att is OhsaiProperty && node.Descendants().Where(s => p.Name == s.Name).Count() > 0
                         select p;

            foreach (var prop in props)
            {
                var propName = prop.Name;
                var nodeValue = node.Descendants().Where(s => propName == s.Name).FirstOrDefault();
                this.SetPropTypeValue(obj, nodeValue.Value, prop);
            }
            foreach (var field in fields)
            {
                var fieldName = field.Name;
                var nodeValue = node.Descendants().Where(s => fieldName == s.Name).FirstOrDefault();
                this.SetFieldTypeValue(obj, nodeValue.Value, field);
            }

            // look for any properties or fields that implement the IAgent interface and are marked with a OhsaiAgentProperty
            // if we find any of these then set there value
            var agentProps = from p in obj.GetType().GetProperties()
                             from att in p.GetCustomAttributes(false)
                             where att is OhsaiAgentProperty && (p.PropertyType == typeof(IAgent) || p.PropertyType.GetInterface(typeof(IAgent).FullName) != null)
                             select p;
            var agentFields = from p in obj.GetType().GetFields()
                              from att in p.GetCustomAttributes(false)
                              where att is OhsaiAgentProperty && (p.FieldType == typeof(IAgent) || p.FieldType.GetInterface(typeof(IAgent).FullName) != null)
                              select p;

            foreach (var prop in agentProps)
                prop.SetValue(obj, agent, null);

            foreach (var field in agentFields)
                field.SetValue(obj, agent);
        }

        private void SetPropTypeValue(object obj, string value, PropertyInfo prop)
        {
            if (prop.PropertyType == typeof(int))
            {
                prop.SetValue(obj, int.Parse(value), null);
            }
            else if (prop.PropertyType == typeof(float))
            {
                prop.SetValue(obj, float.Parse(value), null);
            }
            else if (prop.PropertyType == typeof(bool))
            {
                prop.SetValue(obj, bool.Parse(value), null);
            }
            else if (prop.PropertyType == typeof(string))
            {
                prop.SetValue(obj, value, null);
            }
            else if (prop.PropertyType == typeof(TimeSpan))
            {
                prop.SetValue(obj, TimeSpan.Parse(value), null);
            }
            else if (prop.PropertyType.BaseType == typeof(Enum))
            {
                prop.SetValue(obj, Enum.Parse(prop.PropertyType, value, true), null);
            }
            else
            {
                // unknown type that cannot be parsed so see if a custom reader was specified
                var reader = from r in prop.GetCustomAttributes(false)
                             where r is Ohsai.OhsaiReader
                             select r;

                var item = reader.FirstOrDefault() as OhsaiReader;
                if (item != null)
                {
                    var readerObj = this.FindAndCreateType<IDataReader>(item.Reader);
                    var parsedValue = readerObj.Read(value);
                    prop.SetValue(obj, parsedValue, null);

                }
            }
        }

        private void SetFieldTypeValue(object obj, string value, FieldInfo prop)
        {
            if (prop.FieldType == typeof(int))
            {
                prop.SetValue(obj, int.Parse(value));
            }
            else if (prop.FieldType == typeof(float))
            {
                prop.SetValue(obj, float.Parse(value));
            }
            else if (prop.FieldType == typeof(bool))
            {
                prop.SetValue(obj, bool.Parse(value));
            }
            else if (prop.FieldType == typeof(string))
            {
                prop.SetValue(obj, value);
            }
            else if (prop.FieldType == typeof(TimeSpan))
            {
                prop.SetValue(obj, TimeSpan.Parse(value));
            }
            else if (prop.FieldType.BaseType == typeof(Enum))
            {
                prop.SetValue(obj, Enum.Parse(prop.FieldType, value, true));
            }
            else
            {
                // unknown type that cannot be parsed so see if a custom reader was specified
                var reader = from r in prop.GetCustomAttributes(false)
                             where r is Ohsai.OhsaiReader
                             select r;

                var item = reader.FirstOrDefault() as OhsaiReader;
                if (item != null)
                {
                    var readerObj = this.FindAndCreateType<IDataReader>(item.Reader);
                    var parsedValue = readerObj.Read(value);
                    prop.SetValue(obj, parsedValue);

                }
            }
        }

        //private T FindAndCreateType<T>(string name)
        //{
        //    return this.FindAndCreateType<T>(name, false, null);
        //}

        //private T FindAndCreateType<T>(string name, bool fullname)
        //{
        //    return this.FindAndCreateType<T>(name, fullname, null);
        //}

        //private T FindAndCreateType<T>(string name, Type attType)
        //{
        //    return this.FindAndCreateType<T>(name, false, attType);
        //}

        private T FindAndCreateType<T>(string name)//, bool fullname, object attType)
        {
            //if (attType != null)
            //{
            //    var item = from asm in AppDomain.CurrentDomain.GetAssemblies()
            //               where asm.GlobalAssemblyCache == false
            //               from ty in asm.GetTypes()
            //               where (fullname ? ty.FullName : ty.Name) == name && ty.GetInterface(typeof(T).FullName) != null &&
            //                      ty.GetCustomAttributes(true).Where(s => s.GetType().Equals( attType)).Count() > 0
            //               select ty;

            //    var respType = item.FirstOrDefault();
            //    var obj = respType.Assembly.CreateInstance(respType.FullName);
            //    return (T)obj;
            //}
            //else
            //{
            var item = from asm in AppDomain.CurrentDomain.GetAssemblies()
                       where asm.GlobalAssemblyCache == false
                       from ty in asm.GetTypes()
                       // where (fullname ? ty.FullName : ty.Name) == name && ty.GetInterface(typeof(T).FullName) != null
                       where ty.FullName == name && ty.GetInterface(typeof(T).FullName) != null
                       select ty;

            var respType = item.FirstOrDefault();
            var obj = respType.Assembly.CreateInstance(respType.FullName);
            return (T)obj;
            // }
        }
    }
}
