﻿#if XNA
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;


namespace Ohsai.Sensors
{
    /// <summary>
    /// Provides a enum for specifying the XBox360 gamepad source input.
    /// </summary>
    public enum GamePadSource
    {
        /// <summary>
        /// The gamepad's Big button.
        /// </summary>
        ButtonBig,
        /// <summary>
        /// The gamepad's A button.
        /// </summary>
        ButtonA,
        /// <summary>
        /// The gamepad's B button.
        /// </summary>
        ButtonB,
        /// <summary>
        /// The gamepad's X button.
        /// </summary>
        ButtonX,
        /// <summary>
        /// The gamepad's Y button.
        /// </summary>
        ButtonY,
        /// <summary>
        /// The gamepad's Left trigger button.
        /// </summary>
        LeftTrigger,
        /// </summary>
        RightTrigger,
        /// <summary>
        /// The gamepad's Left sholder button.
        /// </summary>
        LeftSholder,
        /// <summary>
        /// The gamepad's Right sholder button.
        /// </summary>
        RightSholder,
        /// <summary>
        /// The gamepad's left stick position.
        /// </summary>
        LeftStickPosition,
        /// <summary>
        /// The gamepad's right stick position.
        /// </summary>
        RightStickPosition,
        /// <summary>
        /// The gamepad's X axis value of the left control stick.
        /// </summary>
        LeftStickXPosition,
        /// <summary>
        /// The gamepad's X axis value of the right control stick.
        /// </summary>
        RightStickXPosition,
        /// <summary>
        /// The gamepad's Y axis value of the left control stick.
        /// </summary>
        LeftStickYPosition,
        /// <summary>
        /// The gamepad's Y axis value of the right control stick.
        /// </summary>
        RightStickYPosition,
        /// <summary>
        /// The gamepad's left control stick button.
        /// </summary>
        LeftStickButton,
        /// <summary>
        /// The gamepad's right control stick button.
        /// </summary>
        RightStickButton,
        /// <summary>
        /// The up state for the gamepad's DPad.
        /// </summary>
        DPadUp,
        /// <summary>
        /// The down state for the gamepad's DPad.
        /// </summary>
        DPadDown,
        /// <summary>
        /// The left state for the gamepad's DPad.
        /// </summary>
        DPadLeft,
        /// <summary>
        /// The right state for the gamepad's DPad.
        /// </summary>
        DPadRight,
        /// <summary>
        /// The up state for the gamepad's DPad.
        /// </summary>
        Up = DPadUp,
        /// <summary>
        /// The down state for the gamepad's DPad.
        /// </summary>
        Down = DPadDown,
        /// <summary>
        /// The left state for the gamepad's DPad.
        /// </summary>
        Left = DPadLeft,
        /// <summary>
        /// The right state for the gamepad's DPad.
        /// </summary>
        Right = DPadRight,
        /// <summary>
        /// The gamepad's back button.
        /// </summary>
        BackButton,
        /// <summary>
        /// The gamepad's start button.
        /// </summary>
        StartButton,
        /// <summary>
        /// The gamepad's back button.
        /// </summary>
        Back = BackButton,
        /// <summary>
        /// The gamepad's start button.
        /// </summary>
        Start = StartButton,
        /// <summary>
        /// The gamepad's back button.
        /// </summary>
        ButtonBack = BackButton,
        /// <summary>
        /// The gamepad's start button.
        /// </summary>
        ButtonStart = StartButton
    }

    public struct GamePadItem
    {
        public GamePadSource Source { get; set; }
        public PlayerIndex Player { get; set; }
        public float Value { get; set; }
        public Vector3 Position { get; set; }

        public GamePadItem(GamePadSource source, PlayerIndex player, Vector2 position)
            : this()
        {
            this.Source = source;
            this.Player = player;
            this.Position = new Vector3(position.X, position.Y, 0);
        }

        public GamePadItem(GamePadSource source, PlayerIndex player, float value)
            : this()
        {
            this.Source = source;
            this.Player = player;
            this.Value = value;
        }
    }

    [OhsaiSensor(Name = "GamePadSensor", Description = "Provides a sensor for gamepad input.")]
    public class GamePadSensor : BaseSensor
    {
        [OhsaiProperty(Name = "Player", Description = "Gets or sets the player to detect input from.")]
        public PlayerIndex Player { get; set; }
        [OhsaiProperty(Name = "Source", Description = "Gets or sets the game pad input source.")]
        public GamePadSource Source { get; set; }


        public GamePadSensor()
        {
            this.Player = PlayerIndex.One;
        }

        public override void Update(object sender, EventArgs e)
        {
            this.Results.Clear();
            var gamepad = GamePad.GetState(this.Player);
            switch (this.Source)
            {
                case GamePadSource.ButtonBig:
                    if (gamepad.Buttons.BigButton == ButtonState.Pressed) this.Results.Add(new GamePadItem(this.Source, this.Player, 1));
                    break;
                case GamePadSource.ButtonA:
                    if (gamepad.Buttons.A == ButtonState.Pressed) this.Results.Add(new GamePadItem(this.Source, this.Player, 1));
                    break;
                case GamePadSource.ButtonB:
                    if (gamepad.Buttons.B == ButtonState.Pressed) this.Results.Add(new GamePadItem(this.Source, this.Player, 1));
                    break;
                case GamePadSource.ButtonX:
                    if (gamepad.Buttons.X == ButtonState.Pressed) this.Results.Add(new GamePadItem(this.Source, this.Player, 1));
                    break;
                case GamePadSource.ButtonY:
                    if (gamepad.Buttons.Y == ButtonState.Pressed) this.Results.Add(new GamePadItem(this.Source, this.Player, 1));
                    break;
                case GamePadSource.LeftTrigger:
                    if (gamepad.Triggers.Left != 0) this.Results.Add(new GamePadItem(this.Source, this.Player, gamepad.Triggers.Left));
                    break;
                case GamePadSource.RightTrigger:
                    if (gamepad.Triggers.Right != 0) this.Results.Add(new GamePadItem(this.Source, this.Player, gamepad.Triggers.Right));
                    break;
                case GamePadSource.LeftSholder:
                    if (gamepad.Buttons.LeftShoulder == ButtonState.Pressed) this.Results.Add(new GamePadItem(this.Source, this.Player, 1));
                    break;
                case GamePadSource.RightSholder:
                    if (gamepad.Buttons.RightShoulder == ButtonState.Pressed) this.Results.Add(new GamePadItem(this.Source, this.Player, 1));
                    break;
                case GamePadSource.LeftStickXPosition:
                    if (gamepad.ThumbSticks.Left.X != 0) this.Results.Add(new GamePadItem(this.Source, this.Player, gamepad.ThumbSticks.Left.X));
                    break;
                case GamePadSource.RightStickXPosition:
                    if (gamepad.ThumbSticks.Right.X != 0) this.Results.Add(new GamePadItem(this.Source, this.Player, gamepad.ThumbSticks.Right.X));
                    break;
                case GamePadSource.LeftStickYPosition:
                    if (gamepad.ThumbSticks.Left.Y != 0) this.Results.Add(new GamePadItem(this.Source, this.Player, gamepad.ThumbSticks.Left.Y));
                    break;
                case GamePadSource.RightStickYPosition:
                    if (gamepad.ThumbSticks.Right.Y != 0) this.Results.Add(new GamePadItem(this.Source, this.Player, gamepad.ThumbSticks.Right.Y));
                    break;
                case GamePadSource.LeftStickPosition:
                    if (gamepad.ThumbSticks.Left != Vector2.Zero) this.Results.Add(new GamePadItem(this.Source, this.Player, gamepad.ThumbSticks.Left));
                    break;
                case GamePadSource.RightStickPosition:
                    if (gamepad.ThumbSticks.Right != Vector2.Zero) this.Results.Add(new GamePadItem(this.Source, this.Player, gamepad.ThumbSticks.Right));
                    break;
                case GamePadSource.LeftStickButton:
                    if (gamepad.Buttons.LeftStick == ButtonState.Pressed) this.Results.Add(new GamePadItem(this.Source, this.Player, 1));
                    break;
                case GamePadSource.RightStickButton:
                    if (gamepad.Buttons.RightStick == ButtonState.Pressed) this.Results.Add(new GamePadItem(this.Source, this.Player, 1));
                    break;
                case GamePadSource.DPadUp:
                    if (gamepad.DPad.Up == ButtonState.Pressed) this.Results.Add(new GamePadItem(this.Source, this.Player, 1));
                    break;
                case GamePadSource.DPadDown:
                    if (gamepad.DPad.Down == ButtonState.Pressed) this.Results.Add(new GamePadItem(this.Source, this.Player, 1));
                    break;
                case GamePadSource.DPadLeft:
                    if (gamepad.DPad.Left == ButtonState.Pressed) this.Results.Add(new GamePadItem(this.Source, this.Player, 1));
                    break;
                case GamePadSource.DPadRight:
                    if (gamepad.DPad.Right == ButtonState.Pressed) this.Results.Add(new GamePadItem(this.Source, this.Player, 1));
                    break;
                case GamePadSource.BackButton:
                    if (gamepad.Buttons.Back == ButtonState.Pressed) this.Results.Add(new GamePadItem(this.Source, this.Player, 1));
                    break;
                case GamePadSource.StartButton:
                    if (gamepad.Buttons.Start == ButtonState.Pressed) this.Results.Add(new GamePadItem(this.Source, this.Player, 1));
                    break;
                default:
                    throw new ArgumentOutOfRangeException();
            }
        }
    }
}
#endif